/*
 * BottomUpHierarchicalAligner.h
 *
 *  Created on: 17.01.2013
 *      Author: kaeshammer
 */

#ifndef EXTRACT_LCFRS_HIERARCHICALALIGNER_H_
#define EXTRACT_LCFRS_HIERARCHICALALIGNER_H_

#include "TranslationUnitStore.h"

#include <list>
#include <queue>
#include <map>
#include <vector>
#include <boost/dynamic_bitset.hpp>

namespace Moses {
namespace LCFRS {

typedef boost::dynamic_bitset<> boost_bitset;
typedef std::pair<std::string, std::string> BiLabel;
typedef std::pair<boost_bitset, boost_bitset> BiBitset;

class MonoItem {
public:
  std::string l;
  boost::dynamic_bitset<> bs;
  MonoItem(std::string label, const boost_bitset & bitset)
      : l(label), bs(bitset){};
  ~MonoItem(){};
  std::string toString();
};

class BiItem {
public:
  MonoItem itemSource;
  MonoItem itemTarget;
  BiItem* back1;
  BiItem* back2;
  BiItem* back3;
  std::vector<BiItem*> equiItems;
  bool status;
  BiItem(std::string labelSource, const boost_bitset & bitsetSource,
      std::string labelTarget, const boost_bitset & bitsetTarget)
      : itemSource(labelSource, bitsetSource), itemTarget(labelTarget, bitsetTarget),
        back1(NULL), back2(NULL), back3(NULL), status(false){};
  BiItem(std::string labelSource, const boost_bitset & bitsetSource,
          std::string labelTarget, const boost_bitset & bitsetTarget,
          BiItem* b1)
          : itemSource(labelSource, bitsetSource), itemTarget(labelTarget, bitsetTarget),
            back1(b1), back2(NULL), back3(NULL), status(false){};
  BiItem(std::string labelSource, const boost_bitset & bitsetSource,
        std::string labelTarget, const boost_bitset & bitsetTarget,
        BiItem* b1, BiItem* b2)
        : itemSource(labelSource, bitsetSource), itemTarget(labelTarget, bitsetTarget),
          back1(b1), back2(b2), back3(NULL), status(false){};
  BiItem(std::string labelSource, const boost_bitset & bitsetSource,
        std::string labelTarget, const boost_bitset & bitsetTarget,
        BiItem* b1, BiItem* b2, BiItem* b3)
        : itemSource(labelSource, bitsetSource), itemTarget(labelTarget, bitsetTarget),
          back1(b1), back2(b2), back3(b3), status(false){};
  ~BiItem(){};
  std::string toString();
};



std::ostream& operator<<(std::ostream& os, const MonoItem& item);
std::ostream& operator<<(std::ostream& os, const BiItem& item);
std::ostream& operator<<(std::ostream& os, const BiBitset& bs);

//typedef std::map<BiBitset,  std::vector<BiItem*> > my_inner_chart;
typedef std::map<BiBitset, BiItem*> my_inner_chart;
typedef std::map<BiLabel, my_inner_chart > my_chart;


class BiChart {
private:

public:
  my_chart chart;

  BiChart() {};
  ~BiChart() {};
  void addItem(BiItem* item);
  void addItem(BiItem* item, bool& alreadyIn);
  bool equiItemExists(BiItem*item);
  my_chart::const_iterator get(const BiLabel& label);
  friend std::ostream& operator<<(std::ostream& os, const BiChart& bichart);
};

class DerivationNode { // TODO
public:
  BiItem* data;
  DerivationNode* child1;
  DerivationNode* child2;
  DerivationNode(BiItem* data, DerivationNode* c1, DerivationNode* c2)
    :data(data), child1(c1), child2(c2) {};
};

class Derivation { // TODO
public:
  DerivationNode* root;
  Derivation(DerivationNode* root) : root(root){};
};

typedef std::map<int, int> IntMap;

class BottomUpHierarchicalAligner {
private:
  static const std::string NONTERM;
  static const std::string PRETERM;
  TranslationUnitStore m_tuStore;
  std::list<BiItem*> agenda;
  BiChart m_chart;
  std::vector<BiItem*> goals;
  std::vector<Derivation*> derivations; // TODO
  unsigned int m_maxNumBlocksSource;
  unsigned int m_maxNumBlocksTarget;
  bool m_normalForm;
  bool validateHierAlignment();
  bool doParse(unsigned int kS, unsigned int kT);
  void printTrace(BiItem* item, int indent);
public:
  BottomUpHierarchicalAligner(const TranslationUnitStore &tuStore, unsigned int kS, unsigned int kT, bool nf)
      :m_tuStore(tuStore), m_maxNumBlocksSource(kS), m_maxNumBlocksTarget(kT), m_normalForm(nf){ validateHierAlignment(); };
  ~BottomUpHierarchicalAligner();
  bool success();
  void printChart();
  void printOneDerivation();
  IntMap in2outSource;
  IntMap out2inSource;
  IntMap in2outTarget;
  IntMap out2inTarget;
};



}
}

#endif /* EXTRACT_LCFRS_HIERARCHICALALIGNER_H_ */
